from typing import List, Tuple
import numpy as np
from tdw.controller import Controller
from tdw.tdw_utils import TDWUtils
from tdw.add_ons.benchmark import Benchmark


class PerformanceBenchmarkController(Controller):
    """
    Use this controller to create performance benchmarks.
    """
    def __init__(self, port: int = 1071, check_version: bool = True, launch_build: bool = True):
        super().__init__(port=port, check_version=check_version, launch_build=launch_build)
        """:field
        The [`Benchmark`](../add_ons/benchmark.md) add-on.
        """
        self.benchmark: Benchmark = Benchmark()
        self.add_ons.append(self.benchmark)

    def run(self, boxes: bool = False, render_quality: int = 0, post_processing: bool = False,
            pass_masks: List[str] = None, png: bool = False, transforms: bool = False, rigidbodies: bool = False,
            collisions: bool = False, bounds: bool = False, screen_size: int = 256, junk: int = None,
            images: bool = False, id_colors: bool = False, collision_enter: bool = True, collision_exit: bool = True,
            collision_stay: bool = True, env_collisions: bool = True, occlusion: bool = False,
            num_frames: int = 5000) -> float:
        """
        Run a performance benchmark test.

        :param boxes: If True, add 100 cubes to the scene.
        :param render_quality: The render quality (0 to 5).
        :param post_processing: If True, enable post-processing.
        :param pass_masks: The pass masks.
        :param png: If True, encode the `_img` pass as a .png. If False, encode the `_img` pass as a .jpg. Ignored if `passes` doesn't include `_img` or if `images` is `False`.
        :param transforms: If True, send `Transforms` data per frame.
        :param rigidbodies: If True, send `Rigidbodies` data per frame.
        :param collisions: If True, send collision data per frame.
        :param bounds: If True, send `Bounds` data per frame.
        :param screen_size: The screen size in pixels.
        :param junk: If not None, request this much junk data per frame.
        :param images: If True, add an avatar and send `Images` data per frame.
        :param id_colors: If True, add an avatar and send `IdPassSegmentationColors` data per frame.
        :param collision_enter: If `collisions` is `True`, listen for collision enter events.
        :param collision_exit: If `collisions` is `True`, listen for collision exit events.
        :param collision_stay: If `collisions` is `True`, listen for collision stay events.
        :param env_collisions: If `collisions` is `True`, listen for environment collision events.
        :param occlusion: If True, add an avatar and send `Occlusion` data per frame.
        :param num_frames: Iterate for this many frames.

        :return: The frames per second (FPS).
        """

        commands = [{'$type': "load_scene",
                     'scene_name': "ProcGenScene"},
                    {"$type": "set_render_quality",
                     "render_quality": render_quality},
                    {"$type": "set_post_process",
                     "value": post_processing},
                    {"$type": "set_screen_size",
                     "width": screen_size,
                     "height": screen_size},
                    {"$type": "set_img_pass_encoding",
                     "value": png},
                    TDWUtils.create_empty_room(30, 30)]
        if boxes:
            y = 0.5
            # Create 100 boxes.
            object_ids = []
            for i in range(100):
                commands.append({"$type": "load_primitive_from_resources",
                                 "primitive_type": "Cube",
                                 "id": i,
                                 "position": {"x": 0, "y": y, "z": 0},
                                 "orientation": {"x": 0, "y": 0, "z": 0}})
                object_ids.append(i)
                y += 1.5
        transforms = "always" if transforms else "never"
        rigidbodies = "always" if rigidbodies else "never"
        bounds = "always" if bounds else "never"
        commands.extend([{"$type": "send_transforms",
                          "frequency": transforms},
                         {"$type": "send_rigidbodies",
                          "frequency": rigidbodies},
                         {"$type": "send_bounds",
                          "frequency": bounds}])
        if collisions:
            collision_types = ["obj"]
            # Listen for environment collisions, too.
            if env_collisions:
                collision_types.append("env")
            commands.append({"$type": "send_collisions",
                             "enter": collision_enter,
                             "stay": collision_stay,
                             "exit": collision_exit,
                             "collision_types": collision_types})
        if junk is not None:
            commands.append({"$type": "send_junk",
                             "frequency": "always",
                             "length": junk})
        if images or id_colors or occlusion:
            commands.extend([{"$type": "create_avatar",
                             "id": "a",
                              "type": "A_Img_Caps_Kinematic"},
                             {"$type": "teleport_avatar_to",
                              "position": {"x": 0, "y": 1.5, "z": -2}}])
        if images and pass_masks is not None:
            commands.extend([{"$type": "set_pass_masks",
                              "pass_masks": pass_masks},
                             {"$type": "send_images",
                              "frequency": "always"}])
        if id_colors:
            commands.append({"$type": "send_id_pass_segmentation_colors",
                             "frequency": "always"})
        if occlusion:
            commands.append({"$type": "send_occlusion",
                             "frequency": "always"})
        self.communicate(commands)
        frame = 0
        self.benchmark.start()
        while frame < num_frames:
            self.communicate([])
            frame += 1
        self.benchmark.stop()
        return round(self.benchmark.fps)

    def agent_benchmark(self, random_seed: int = None) -> float:
        """
        Start a performance benchmark using an embodied agent.

        :param random_seed: The random seed. If None, a random seed is randomly selected.

        :return: The frames per second (FPS).
        """

        def __get_position() -> Tuple[float, float]:
            px, pz = free_cells[free_cell_indices.pop(0)]
            return (rng.uniform(px - cell_size * 0.33, px + cell_size * 0.33),
                    rng.uniform(pz - cell_size * 0.33, pz + cell_size * 0.33))

        if random_seed is None:
            rng: np.random.RandomState = np.random.RandomState()
        else:
            rng = np.random.RandomState(random_seed)
        cell_size = 0.5
        occupancy_map = np.array([[-1, -1, -1, -1, -1, -1, -1, -1, -1],
                                  [-1, 0, 0, 0, 0, 0, 0, 0, -1],
                                  [-1, 0, 0, 0, 0, 0, 0, 0, -1],
                                  [-1, 0, 0, 0, 0, 0, 0, 0, -1],
                                  [-1, 0, 0, 0, 0, 0, 0, 0, -1],
                                  [-1, 0, 0, 0, 0, 0, 0, 0, -1],
                                  [-1, 0, 0, 0, 0, 0, 0, 0, -1],
                                  [-1, 0, 0, 0, 0, 0, 0, 0, -1],
                                  [-1, -1, -1, -1, -1, -1, -1, -1, -1]])
        free_cells = []
        for ix, iz in np.ndindex(occupancy_map.shape):
            if occupancy_map[ix][iz] == 0:
                x = -2.0 + (ix * cell_size)
                z = -2.0 + (iz * cell_size)
                free_cells.append((x, z))
        free_cell_indices = list(range(len(free_cells)))
        rng.shuffle(free_cell_indices)
        free_cell_indices = list(free_cell_indices)

        commands = [{'$type': "load_scene",
                     'scene_name': "ProcGenScene"},
                    {"$type": "set_render_quality",
                     "render_quality": 0},
                    {"$type": "set_screen_size",
                     "width": 256,
                     "height": 256},
                    {"$type": "set_img_pass_encoding",
                     "value": False},
                    {"$type": "set_post_process",
                     "value": False},
                    TDWUtils.create_empty_room(5, 5)]
        # Add objects.
        for model_name in ["bastone_floor_lamp", "arturoalvarez_v_floor_lamp"]:
            for i in range(4):
                object_x, object_z = __get_position()
                commands.extend(self.get_add_physics_object(model_name=model_name,
                                                            object_id=self.get_unique_id(),
                                                            position={"x": object_x, "y": 0, "z": object_z}))
        # Add the avatar.
        avatar_x, avatar_z = __get_position()
        commands.extend([{"$type": "create_avatar",
                          "type": "A_Img_Caps",
                          "id": "a"},
                         {"$type": "teleport_avatar_to",
                          "position": {"x": avatar_x, "y": 0, "z": avatar_z}},
                         {"$type": "set_pass_masks",
                          "pass_masks": ["_img"]},
                         {"$type": "send_id_pass_segmentation_colors",
                          "frequency": "always"},
                         {"$type": "send_transforms",
                          "frequency": "always"},
                         {"$type": "send_collisions",
                          "frequency": "always"},
                         {"$type": "send_rigidbodies",
                          "frequency": "always"},
                         {"$type": "send_images",
                          "frequency": "always"}])
        # Start the benchmark.
        self.communicate(commands)
        self.benchmark.start()
        for trial_frame in range(1000):
            commands.clear()
            random_action = np.random.uniform(-200, 200, 2)
            # Actual TDW actions.
            for p_idx in range(5):
                commands.extend([{"$type": "move_avatar_forward_by",
                                 "magnitude": random_action[0]},
                                 {"$type": "rotate_avatar_by",
                                  "angle": random_action[1] * 8,
                                  "axis": "yaw"},
                                 {"$type": "step_physics",
                                  "frames": 1}])
            self.communicate(commands)
        self.benchmark.stop()
        return self.benchmark.fps

    def flex_benchmark(self) -> float:
        """
        Start a Flex performance benchmark.

        :return: The frames per second (FPS).
        """

        self.communicate([{'$type': 'load_scene',
                           'scene_name': 'ProcGenScene'},
                          {'$type': 'create_exterior_walls',
                           'walls': [{'x': 0, 'y': 0}, {'x': 0, 'y': 1}, {'x': 0, 'y': 2}, {'x': 0, 'y': 3},
                                     {'x': 0, 'y': 4}, {'x': 1, 'y': 0}, {'x': 1, 'y': 4}, {'x': 2, 'y': 0},
                                     {'x': 2, 'y': 4}, {'x': 3, 'y': 0}, {'x': 3, 'y': 4}, {'x': 4, 'y': 0},
                                     {'x': 4, 'y': 1}, {'x': 4, 'y': 2}, {'x': 4, 'y': 3}, {'x': 4, 'y': 4}]},
                          {'$type': 'create_proc_gen_ceiling'},
                          {'$type': 'convexify_proc_gen_room'},
                          {'$type': 'create_avatar',
                           'id': 'a',
                           'type': 'A_Img_Caps_Kinematic'},
                          {'$type': 'teleport_avatar_to',
                           'position': {'x': -2, 'y': 1, 'z': -2}},
                          {'$type': 'set_avatar_forward',
                           'forward': {'x': 2, 'y': -1, 'z': 2}},
                          {'$type': 'send_camera_matrices',
                           'frequency': 'always'},
                          {'$type': 'send_flex_particles',
                           'frequency': 'always'},
                          {'$type': 'create_flex_container',
                           'gravity': {'x': 0, 'y': -9.81, 'z': 0},
                           'radius': 0.1875,
                           'solid_rest': 0.125,
                           'fluid_rest': 0.115,
                           'static_friction': 0.5,
                           'dynamic_friction': 0.33333333,
                           'particle_friction': 0.166666666,
                           'substep_count': 8,
                           'iteration_count': 8,
                           'collision_distance': 0.0625,
                           'damping': 0.1,
                           'planes': [{'x': 0, 'y': 1, 'z': 0, 'w': 0}, {'x': 0, 'y': -1, 'z': 0, 'w': 5},
                                      {'x': 1, 'y': 0, 'z': 0, 'w': 2.2}, {'x': -1, 'y': 0, 'z': 0, 'w': 2.2},
                                      {'x': 0, 'y': 0, 'z': 1, 'w': 2.2}, {'x': 0, 'y': 0, 'z': -1, 'w': 2.2}]},
                          {'$type': 'set_time_step',
                           'time_step': 0.05},
                          {'$type': 'set_render_quality',
                           'render_quality': 0},
                          {'$type': 'set_post_process',
                           'value': False},
                          {'$type': 'load_primitive_from_resources',
                           'id': 1,
                           'primitive_type': 'Cube',
                           'position': {'x': 0.16403958415021425, 'y': 0.5186866147245053, 'z': 0.5172312486447144},
                           'orientation': {'x': 0, 'y': 0, 'z': 0}},
                          {"$type": "rotate_object_to",
                           "id": 1,
                           "rotation": {'x': -0.3698290892432926, 'y': 0.09343813858449268, 'z': -0.2922975854616724, 'w': 0.8769594520504459}},
                          {"$type": "scale_object",
                           "id": 1,
                           "scale_factor": {"x": 0.5, "y": 0.5, "z": 0.5}},
                          {'$type': 'set_kinematic_state',
                           'id': 1,
                           'is_kinematic': False},
                          {'$type': 'set_flex_soft_actor',
                           'id': 1,
                           'draw_particles': False,
                           'particle_spacing': 0.125,
                           'cluster_stiffness': 0.21856892364500366},
                          {'$type': 'assign_flex_container',
                           'id': 1,
                           'container_id': 0},
                          {'$type': 'set_flex_particles_mass',
                           'id': 1,
                           'mass': 15.625},
                          {'$type': 'set_color',
                           'id': 1,
                           'color': {'r': 0.36824153984054797, 'g': 0.9571551589530464, 'b': 0.14035078041264515, 'a': 1}},
                          {'$type': 'load_primitive_from_resources',
                           'id': 2,
                           'primitive_type': 'Cube',
                           'position': {'x': 0.47383635425791626, 'y': 0.35827517721218594, 'z': -0.7295636531890959},
                           'orientation': {'x': 0, 'y': 0, 'z': 0}},
                          {"$type": "rotate_object_to",
                           "id": 2,
                           "rotation": {'x': -0.26294416753798533, 'y': -0.28764056724783615, 'z': -0.32404784600690667, 'w': 0.862041914485243}},
                          {"$type": "scale_object",
                           "id": 2,
                           "scale_factor": {"x": 0.5, "y": 0.5, "z": 0.5}},
                          {'$type': 'set_kinematic_state',
                           'id': 2,
                           'is_kinematic': False},
                          {'$type': 'set_flex_soft_actor',
                           'id': 2,
                           'draw_particles': False,
                           'particle_spacing': 0.125,
                           'cluster_stiffness': 0.22055267521432875},
                          {'$type': 'assign_flex_container',
                           'id': 2,
                           'container_id': 0},
                          {'$type': 'set_flex_particles_mass',
                           'id': 2,
                           'mass': 15.625},
                          {'$type': 'set_color',
                           'id': 2,
                           'color': {'r': 0.8700872583584364, 'g': 0.4736080452737105, 'b': 0.8009107519796442, 'a': 1}},
                          {'$type': 'load_primitive_from_resources',
                           'id': 3,
                           'primitive_type': 'Cube',
                           'position': {'x': -0.8337750147387952, 'y': 0.3888592806405162, 'z': -0.9812865902869348},
                           'orientation': {'x': 0, 'y': 0, 'z': 0}},
                          {"$type": "rotate_object_to",
                           "id": 3,
                           "rotation": {'x': 0.2648241162596214, 'y': 0.23805685090753817, 'z': 0.263751099599952, 'w': 0.896455509572624}},
                          {"$type": "scale_object",
                           "id": 3,
                           "scale_factor": {"x": 0.5, "y": 0.5, "z": 0.5}},
                          {'$type': 'set_kinematic_state',
                           'id': 3,
                           'is_kinematic': False},
                          {'$type': 'set_flex_soft_actor',
                           'id': 3,
                           'draw_particles': False,
                           'particle_spacing': 0.125,
                           'cluster_stiffness': 0.5247127393571944},
                          {'$type': 'assign_flex_container',
                           'id': 3,
                           'container_id': 0},
                          {'$type': 'set_flex_particles_mass',
                           'id': 3,
                           'mass': 15.625},
                          {'$type': 'set_color',
                           'id': 3,
                           'color': {'r': 0.5204774795512048, 'g': 0.6788795301189603, 'b': 0.7206326547259168, 'a': 1}},
                          {'$type': 'send_collisions',
                           'frequency': 'always',
                           'enter': True,
                           'exit': False,
                           'stay': False},
                          {'$type': 'send_transforms',
                           'frequency': 'always'}])
        self.benchmark.start()
        for i in range(2000):
            self.communicate([])
        self.benchmark.stop()
        return self.benchmark.fps



from typing import List, Dict
from tdw.add_ons.audio_initializer_base import AudioInitializerBase
from tdw.physics_audio.audio_material import AudioMaterial


class ResonanceAudioInitializer(AudioInitializerBase):
    """
    Initialize Resonance Audio.

    This assumes that an avatar corresponding to `avatar_id` has already been added to the scene.
    """

    """:class_var
    A dictionary. Key = A Resonance Audio material string. Value = An [`AudioMaterial`](../physics_audio/audio_material.md).
    """
    AUDIO_MATERIALS: Dict[str, AudioMaterial] = {"roughPlaster": AudioMaterial.wood_soft,
                                                 "tile": AudioMaterial.ceramic,
                                                 "concrete": AudioMaterial.ceramic,
                                                 "wood": AudioMaterial.wood_soft,
                                                 "smoothPlaster": AudioMaterial.wood_soft,
                                                 "acousticTile": AudioMaterial.cardboard,
                                                 "glass": AudioMaterial.glass,
                                                 "parquet": AudioMaterial.wood_medium,
                                                 "marble": AudioMaterial.stone,
                                                 "grass": AudioMaterial.fabric,
                                                 "brick": AudioMaterial.ceramic,
                                                 "metal": AudioMaterial.metal}

    def __init__(self, avatar_id: str = "a", region_id: int = -1, floor: str = "parquet", ceiling: str = "acousticTile",
                 front_wall: str = "smoothPlaster", back_wall: str = "smoothPlaster", left_wall: str = "smoothPlaster",
                 right_wall: str = "smoothPlaster", framerate: int = 60):
        """
        :param avatar_id: The ID of the avatar.
        :param region_id: The ID of the scene region (room) to enable reverberation in. If -1, the reverb space will encapsulate the entire scene instead of a single room.
        :param floor: The floor material. [Read this for a list of options.](../../api/command_api.md#set_reverb_space_simple)
        :param ceiling: The ceiling material. [Read this for a list of options.](../../api/command_api.md#set_reverb_space_simple)
        :param front_wall: The front wall material. [Read this for a list of options.](../../api/command_api.md#set_reverb_space_simple)
        :param back_wall: The back wall material. [Read this for a list of options.](../../api/command_api.md#set_reverb_space_simple)
        :param left_wall: The left wall material. [Read this for a list of options.](../../api/command_api.md#set_reverb_space_simple)
        :param right_wall: The right wall material. [Read this for a list of options.](../../api/command_api.md#set_reverb_space_simple)
        :param framerate: The target simulation framerate.
        """

        super().__init__(avatar_id=avatar_id, framerate=framerate)
        """:field
        The ID of the scene region (room) to enable reverberation in. If -1, the reverb space will encapsulate the entire scene instead of a single room.
        """
        self.region_id = region_id
        """:field
        The floor material. [Read this for a list of options.](../../api/command_api.md#set_reverb_space_simple)
        """
        self.floor: str = floor
        """:field
        The ceiling material. [Read this for a list of options.](../../api/command_api.md#set_reverb_space_simple)
        """
        self.ceiling: str = ceiling
        """:field
        The front wall material. [Read this for a list of options.](../../api/command_api.md#set_reverb_space_simple)
        """
        self.front_wall: str = front_wall
        """:field
        The back wall material. [Read this for a list of options.](../../api/command_api.md#set_reverb_space_simple)
        """
        self.back_wall: str = back_wall
        """:field
        The left wall material. [Read this for a list of options.](../../api/command_api.md#set_reverb_space_simple)
        """
        self.left_wall: str = left_wall
        """:field
        The right wall material. [Read this for a list of options.](../../api/command_api.md#set_reverb_space_simple)
        """
        self.right_wall: str = right_wall

    def get_initialization_commands(self) -> List[dict]:
        commands = super().get_initialization_commands()
        commands.insert(0, {"$type": "set_reverb_space_simple",
                            "region_id": self.region_id,
                            "reverb_floor_material": self.floor,
                            "reverb_ceiling_material": self.ceiling,
                            "reverb_front_wall_material": self.front_wall,
                            "reverb_back_wall_material": self.back_wall,
                            "reverb_left_wall_material": self.left_wall,
                            "reverb_right_wall_material": self.right_wall})
        return commands

    def _get_sensor_command_name(self) -> str:
        return "add_environ_audio_sensor"

    def _get_play_audio_command_name(self) -> str:
        return "play_point_source_data"



from typing import List, Dict, Union, Optional
from tdw.add_ons.third_person_camera_base import ThirdPersonCameraBase


class ThirdPersonCamera(ThirdPersonCameraBase):
    """
    Add a third-person camera to the scene. This includes initialization parameters (position, rotation, etc.) and some basic movement parameters (whether to follow or look at a target),.

    ```python
    from tdw.controller import Controller
    from tdw.tdw_utils import TDWUtils
    from tdw.add_ons.third_person_camera import ThirdPersonCamera

    c = Controller(launch_build=False)
    cam = ThirdPersonCamera(avatar_id="c",
                            position={"x": 1, "y": 2.2, "z": -0.5},
                            rotation={"x": 0, "y": -45, "z": 0})
    c.add_ons.append(cam)
    c.communicate(TDWUtils.create_empty_room(12, 12))
    ```

    By itself, a `ThirdPersonCamera` won't capture images (though it will render them on the screen). For image capture, include an `ImageCapture` add-on:

    ```python
    from tdw.controller import Controller
    from tdw.tdw_utils import TDWUtils
    from tdw.add_ons.third_person_camera import ThirdPersonCamera
    from tdw.add_ons.image_capture import ImageCapture

    c = Controller(launch_build=False)
    cam = ThirdPersonCamera(avatar_id="c",
                            position={"x": 1, "y": 2.2, "z": -0.5},
                            rotation={"x": 0, "y": -45, "z": 0})
    cap = ImageCapture(path="images", avatar_ids=["c"])
    c.add_ons.append(cam)
    c.add_ons.append(cap)
    c.communicate(TDWUtils.create_empty_room(12, 12))
    ```

    ## Third-person cameras and avatars

    The `ThirdPersonCamera` is a wrapper class for a standard `A_Img_Caps_Kinematic` TDW avatar. All non-physics avatar commands may be sent for this camera.

    In this document, the words "camera" and "avatar" may be used interchangeably.

    ## Multiple cameras

    Unlike most `AddOn` objects, it is possible to add multiple `ThirdPersonCamera`s to the scene:

    ```python
    from tdw.controller import Controller
    from tdw.tdw_utils import TDWUtils
    from tdw.add_ons.third_person_camera import ThirdPersonCamera
    from tdw.add_ons.image_capture import ImageCapture

    c = Controller(launch_build=False)

    # Add two cameras.
    cam_0 = ThirdPersonCamera(avatar_id="c0",
                              position={"x": 1, "y": 2.2, "z": -0.5},
                              rotation={"x": 0, "y": -45, "z": 0})
    cam_1 = ThirdPersonCamera(avatar_id="c1",
                              position={"x": 2, "y": 1, "z": -5})

    # Enable image capture for both cameras.
    cap = ImageCapture(path="images", avatar_ids=["c0", "c1"])

    c.add_ons.extend([cam_0, cam_1, cap])
    c.communicate(TDWUtils.create_empty_room(12, 12))
    ```
    """

    def __init__(self, avatar_id: str = None, position: Dict[str, float] = None, rotation: Dict[str, float] = None,
                 field_of_view: int = None, look_at: Union[int, Dict[str, float]] = None,
                 follow_object: int = None, follow_rotate: bool = False):
        """
        :param avatar_id: The ID of the avatar (camera). If None, a random ID is generated.
        :param position: The initial position of the object.If None, defaults to `{"x": 0, "y": 0, "z": 0}`.
        :param rotation: The initial rotation of the camera. Can be Euler angles (keys are `(x, y, z)`) or a quaternion (keys are `(x, y, z, w)`). If None, defaults to `{"x": 0, "y": 0, "z": 0}`.
        :param look_at: If not None, rotate look at this target every frame. Overrides `rotation`. Can be an int (an object ID) or an `(x, y, z)` dictionary (a position).
        :param field_of_view: If not None, set the field of view.
        :param follow_object: If not None, follow an object per frame. The `position` parameter will be treated as a relative value from the target object rather than worldspace coordinates.
        :param follow_rotate: If True, match the rotation of the object. Ignored if `follow_object` is None.
        """

        super().__init__(avatar_id=avatar_id, position=position, rotation=rotation, field_of_view=field_of_view)

        """:field
        The target object or position that the camera will look at. Can be None (the camera won't look at a target).
        """
        self.look_at_target: Optional[Union[int, Dict[str, float]]] = look_at
        """:field
        The ID of the object the camera will try to follow. Can be None (the camera won't follow an object).
        """
        self.follow_object: Optional[int] = follow_object
        """:field
        If `follow_object` is not None, this determines whether the camera will follow the object's rotation.
        """
        self.follow_rotate: bool = follow_rotate

    def get_initialization_commands(self) -> List[dict]:
        commands = super().get_initialization_commands()
        commands.extend(self._get_look_at_commands())
        return commands

    def on_send(self, resp: List[bytes]) -> None:
        # Look at and focus on a target object or position.
        if self.look_at_target is not None:
            self.commands.extend(self._get_look_at_commands())
        # Follow a target object.
        if self.follow_object is not None and self.position is not None:
            self.commands.append({"$type": "follow_object",
                                  "object_id": self.follow_object,
                                  "position": self.position,
                                  "rotation": self.follow_rotate,
                                  "avatar_id": self.avatar_id})

    def teleport(self, position: Dict[str, float], absolute: bool = True) -> None:
        """
        Teleport the camera to a new position.

        :param position: The new position of the camera.
        :param absolute: If True, `position` is absolute worldspace coordinates. If False, `position` is relative to the avatar's current position.
        """

        if absolute or self.position is None:
            self.position = position
        else:
            self.position = {"x": self.position["x"] + position["x"],
                             "y": self.position["y"] + position["y"],
                             "z": self.position["z"] + position["z"]}
        self.commands.append({"$type": "teleport_avatar_to",
                              "position": self.position,
                              "avatar_id": self.avatar_id})

    def rotate(self, rotation: Dict[str, float]) -> None:
        """
        Rotate the camera.

        :param rotation: Rotate the camera by these angles (in degrees). Keys are `"x"`, `"y"`, `"z"` and correspond to `(pitch, yaw, roll)`.
        """

        for q, axis in zip(["x", "y", "z"], ["pitch", "yaw", "roll"]):
            self.commands.append({"$type": "rotate_sensor_container_by",
                                  "axis": axis,
                                  "angle": rotation[q],
                                  "avatar_id": self.avatar_id})

    def _get_look_at_commands(self) -> List[dict]:
        """
        :return: A command for looking at a target.
        """

        commands = []
        if self.look_at_target is None:
            return commands
        # Look at and focus on the object.
        elif isinstance(self.look_at_target, int):
            commands.extend([{"$type": "look_at",
                              "object_id": self.look_at_target,
                              "use_centroid": True,
                              "avatar_id": self.avatar_id},
                             {"$type": "focus_on_object",
                              "object_id": self.look_at_target,
                              "use_centroid": True,
                              "avatar_id": self.avatar_id}])
        elif isinstance(self.look_at_target, dict):
            commands.append({"$type": "look_at_position",
                             "position": self.look_at_target,
                             "avatar_id": self.avatar_id})
        else:
            raise TypeError(f"Invalid look-at target: {self.look_at_target}")
        return commands



from pkg_resources import resource_filename
from json import loads
from pathlib import Path
from typing import List, Dict
from tdw.physics_audio.scrape_sub_object import ScrapeSubObject
from tdw.physics_audio.scrape_material import ScrapeMaterial
from tdw.physics_audio.audio_material import AudioMaterial


class ScrapeModel:
    """
    Data for a 3D model being used as a PyImpact scrape surface.
    """

    def __init__(self, model_name: str, sub_objects: List[ScrapeSubObject], visual_material: str,
                 audio_material: AudioMaterial, scrape_material: ScrapeMaterial):
        """
        :param model_name: The name of the model.
        :param sub_objects: A list of [sub-objects that will be used as scrape surfaces](scrape_sub_object.md).
        :param visual_material: The name of the new visual material.
        :param audio_material: The [audio material](audio_material.md).
        :param scrape_material: The [scrape material](scrape_material.md).
        """

        """:field
        The name of the model.
        """
        self.model_name: str = model_name
        """:field
        A list of [sub-objects that will be used as scrape surfaces](scrape_sub_object.md).
        """
        self.sub_objects: List[ScrapeSubObject] = sub_objects
        """:field
        The [audio material](audio_material.md).
        """
        self.audio_material: AudioMaterial = audio_material
        """:field
        The [scrape material](scrape_material.md).
        """
        self.scrape_material: ScrapeMaterial = scrape_material
        """:field
        The name of the new visual material.
        """
        self.visual_material: str = visual_material


def __get_default_scrape_models() -> Dict[str, ScrapeModel]:
    """
    :return: A dictionary of default scrape model data.
    """

    scrape_models: Dict[str, ScrapeModel] = dict()
    default_scrape_materials: Dict[str, Dict[str, str]] = loads(
        Path(resource_filename(__name__, f"scrape_materials.json")).read_text())
    default_scrape_models: Dict[str, dict] = loads(
        Path(resource_filename(__name__, f"scrape_models.json")).read_text())
    for model_name in default_scrape_models:
        visual_material = default_scrape_models[model_name]["visual_material"]
        sub_objects: List[ScrapeSubObject] = list()
        for sub_object in default_scrape_models[model_name]["sub_objects"]:
            sub_objects.append(ScrapeSubObject(name=sub_object["name"],
                                               material_index=sub_object["material_index"]))
        audio_material = AudioMaterial[default_scrape_materials[visual_material]["audio_material"]]
        scrape_material = ScrapeMaterial[default_scrape_materials[visual_material]["scrape_material"]]
        scrape_models[model_name] = ScrapeModel(model_name=model_name, sub_objects=sub_objects,
                                                visual_material=visual_material, audio_material=audio_material,
                                                scrape_material=scrape_material)
    return scrape_models


DEFAULT_SCRAPE_MODELS: Dict[str, ScrapeModel] = __get_default_scrape_models()



from typing import List, Dict, Union, Optional
from overrides import final
import numpy as np
from tdw.controller import Controller
from tdw.librarian import RobotLibrarian, RobotRecord
from tdw.tdw_utils import TDWUtils
from tdw.robot_data.robot_static import RobotStatic
from tdw.robot_data.robot_dynamic import RobotDynamic
from tdw.robot_data.joint_type import JointType
from tdw.add_ons.robot_base import RobotBase


class Robot(RobotBase):
    """
    Add a robot to the scene and set joint targets and add joint forces. It has static and dynamic (per-frame) data for each of its joints.

    ```python
    from tdw.controller import Controller
    from tdw.tdw_utils import TDWUtils
    from tdw.add_ons.robot import Robot

    c = Controller()
    # Add a robot.
    robot = Robot(name="ur5",
                  position={"x": -1, "y": 0, "z": 0.5},
                  robot_id=0)
    c.add_ons.append(robot)
    # Initialize the scene.
    c.communicate([{"$type": "load_scene",
                    "scene_name": "ProcGenScene"},
                   TDWUtils.create_empty_room(12, 12)])
    # Set joint targets.
    robot.set_joint_targets({robot.static.joint_ids_by_name["shoulder_link"]: 15,
                             robot.static.joint_ids_by_name["upper_arm_link"]: -45,
                             robot.static.joint_ids_by_name["forearm_link"]: 60})
    # Wait until the robot stops moving.
    while robot.joints_are_moving():
        c.communicate([])
    c.communicate({"$type": "terminate"})
    ```
    """

    def __init__(self, name: str, robot_id: int = 0, position: Dict[str, float] = None, rotation: Dict[str, float] = None,
                 source: Union[RobotLibrarian, RobotRecord, str] = None):
        """
        :param name: The name of the robot.
        :param robot_id: The ID of the robot.
        :param position: The position of the robot. If None, defaults to `{"x": 0, "y": 0, "z": 0}`.
        :param rotation: The rotation of the robot in Euler angles (degrees). If None, defaults to `{"x": 0, "y": 0, "z": 0}`.
        :param source: The source file of the robot. If None: The source will be the URL of the robot record in TDW's built-in [`RobotLibrarian`](../librarian/robot_librarian.md). If `str`: This is a filepath (starts with `file:///`) or a URL (starts with `http://` or `https://`). If `RobotRecord`: the source is the URL in the record. If `RobotLibrarian`: The source is the record in the provided `RobotLibrarian` that matches `name`.
        """

        super().__init__(robot_id=robot_id, position=position, rotation=rotation)
        """:field
        [Static robot data.](../robot_data/robot_static.md)
        """
        self.static: Optional[RobotStatic] = None
        """:field
        [Dynamic robot data.](../robot_data/robot_dynamic.md)
        """
        self.dynamic: Optional[RobotDynamic] = None
        """:field
        The name of the robot.
        """
        self.name: str = name
        """:field
        The URL or filepath of the robot asset bundle.
        """
        self.url: str = ""
        # The initial target positons for the joints.
        self._initial_targets: dict = dict()
        # If True, we've already started to move the joints to their initial position.
        self._set_initial_targets: bool = False
        self._record: Optional[RobotRecord] = None
        if source is None:
            if "robots.json" not in Controller.ROBOT_LIBRARIANS:
                Controller.ROBOT_LIBRARIANS["robots.json"] = RobotLibrarian()
            self._record = Controller.ROBOT_LIBRARIANS["robots.json"].get_record(name)
            self.url = self._record.get_url()
            self._initial_targets = self._record.targets
        elif isinstance(source, RobotLibrarian):
            self._record = source.get_record(name)
            self.url = self._record.get_url()
            self._initial_targets = self._record.targets
        elif isinstance(source, RobotRecord):
            self.url = source.get_url()
            self._initial_targets = source.targets
        elif isinstance(source, str):
            self.url = source
        else:
            raise TypeError("Invalid type for source: " + source)

    def set_joint_targets(self, targets: Dict[int, Union[float, Dict[str, float]]]) -> None:
        """
        Set target angles or positions for a dictionary of joints.

        :param targets: A dictionary of joint targets. Key = The ID of the joint. Value = the targets. For spherical joints, this must be a Vector3 dictionary, for example `{"x": 40, "y": 0, "z": 0}` (angles in degrees). For revolute joints, this must be a float (an angle in degrees). For prismatic joints, this must be a float (a distance in meters).
        """

        for joint_id in targets:
            joint_type = self.static.joints[joint_id].joint_type
            if joint_type == JointType.spherical:
                self.commands.append({"$type": "set_spherical_target",
                                      "target": targets[joint_id],
                                      "joint_id": joint_id,
                                      "id": self.robot_id})
            elif joint_type == JointType.revolute:
                self.commands.append({"$type": "set_revolute_target",
                                      "target": float(targets[joint_id]),
                                      "joint_id": joint_id,
                                      "id": self.robot_id})
            elif joint_type == JointType.prismatic:
                self.commands.append({"$type": "set_prismatic_target",
                                      "target": float(targets[joint_id]),
                                      "joint_id": joint_id,
                                      "id": self.robot_id})
            else:
                raise Exception(f"Cannot set target for joint type {joint_type}")
            self.dynamic.joints[joint_id].moving = True

    def add_joint_forces(self, forces: Dict[int, Union[float, Dict[str, float]]]) -> None:
        """
        Add torques and forces to a dictionary of joints.

        :param forces: A dictionary of joint forces. Key = The ID of the joint. Value = the targets. For spherical joints, this must be a Vector3 dictionary, for example `{"x": 40, "y": 0, "z": 0}` (torques in Newtons). For revolute joints, this must be a float (a torque in Newtons). For prismatic joints, this must be a float (a force in Newtons).
        """

        for joint_id in forces:
            joint_type = self.static.joints[joint_id].joint_type
            if joint_type == JointType.spherical:
                self.commands.append({"$type": "add_torque_to_spherical",
                                      "torque": forces[joint_id],
                                      "joint_id": joint_id,
                                      "id": self.robot_id})
            elif joint_type == JointType.revolute:
                self.commands.append({"$type": "add_torque_to_revolute",
                                      "torque": float(forces[joint_id]),
                                      "joint_id": joint_id,
                                      "id": self.robot_id})
            elif joint_type == JointType.prismatic:
                self.commands.append({"$type": "add_force_to_prismatic",
                                      "force": float(forces[joint_id]),
                                      "joint_id": joint_id,
                                      "id": self.robot_id})
            else:
                raise Exception(f"Cannot apply torque or force to joint type {joint_type}")
            self.dynamic.joints[joint_id].moving = True

    def stop_joints(self, joint_ids: List[int] = None) -> None:
        """
        Stop the joints at their current angles or positions.

        :param joint_ids: A list of joint IDs. If None, stop all joints.
        """

        if joint_ids is None:
            joint_ids = self.static.joints.values()
        for joint_id in joint_ids:
            angles = self.dynamic.joints[joint_id].angles
            joint_type = self.static.joints[joint_id].joint_type
            if joint_type == JointType.fixed_joint:
                continue
            # Set the target to the angle of the first (and only) revolute drive.
            elif joint_type == JointType.revolute:
                self.commands.append({"$type": "set_revolute_target",
                                      "joint_id": joint_id,
                                      "target": float(angles[0]),
                                      "id": self.robot_id})
            # Convert the current prismatic "angle" back into "radians".
            elif joint_type == JointType.prismatic:
                self.commands.append({"$type": "set_prismatic_target",
                                      "joint_id": joint_id,
                                      "target": float(np.radians(angles[0])),
                                      "id": self.robot_id})
            # Set each spherical drive axis.
            elif joint_type == JointType.spherical:
                self.commands.append({"$type": "set_spherical_target",
                                      "target": TDWUtils.array_to_vector3(angles),
                                      "joint_id": joint_id,
                                      "id": self.robot_id})
            else:
                raise Exception(f"Cannot stop joint type {joint_type}")

    @final
    def _get_add_robot_command(self) -> dict:
        return {"$type": "add_robot",
                "id": self.robot_id,
                "position": self.initial_position,
                "rotation": self.initial_rotation,
                "name": self.name,
                "url": self.url}

    def _cache_static_data(self, resp: List[bytes]) -> None:
        self.static = RobotStatic(robot_id=self.robot_id, resp=resp)

    def _set_dynamic_data(self, resp: List[bytes]) -> None:
        dynamic = RobotDynamic(resp=resp, robot_id=self.robot_id, body_parts=self.static.body_parts,
                               previous=self.dynamic)
        self.dynamic = self._set_joints_moving(dynamic)
        # Set the joints to their initial targets.
        if not self._set_initial_targets:
            self._set_initial_targets = True
            targets = dict()
            for joint_name in self._initial_targets:
                targets[self.static.joint_ids_by_name[joint_name]] = self._initial_targets[joint_name]["target"]
            self.set_joint_targets(targets=targets)



import numpy as np
import random
import math
from scipy.spatial import distance
from tdw.output_data import IsOnNavMesh, Images, Bounds
from PIL import Image
import io
import os
from tdw.controller import Controller
from typing import List, Tuple, Dict, Optional, Union
from tdw.librarian import ModelRecord
from pathlib import Path
import boto3
from botocore.exceptions import ProfileNotFound, ClientError
import base64


class TDWUtils:
    """
    Utility functions for controllers.

    Usage:

    ```python
    from tdw.tdw_utils import TDWUtils
    ```
    """

    VECTOR3_ZERO = {"x": 0, "y": 0, "z": 0}

    # Cached values used during point cloud generation.
    __WIDTH: int = -1
    __HEIGHT: int = -1
    __CAM_TO_IMG_MAT: Optional[np.array] = None

    @staticmethod
    def vector3_to_array(vector3: Dict[str, float]) -> np.array:
        """
        Convert a Vector3 object to a numpy array.

        :param vector3: The Vector3 object, e.g. `{"x": 0, "y": 0, "z": 0}`

        :return A numpy array.
        """

        return np.array([vector3["x"], vector3["y"], vector3["z"]])

    @staticmethod
    def array_to_vector3(arr: np.array) -> Dict[str, float]:
        """
        Convert a numpy array to a Vector3.

        :param arr: The numpy array.

        :return A Vector3, e.g. `{"x": 0, "y": 0, "z": 0}`
        """

        return {"x": float(arr[0]), "y": float(arr[1]), "z": float(arr[2])}

    @staticmethod
    def vector4_to_array(vector4: Dict[str, float]) -> np.array:
        """
        Convert a Vector4 to a numpy array.

        :param vector4: The Vector4 object, e.g. `{"x": 0, "y": 0, "z": 0, "w": 0}`

        :return A numpy array.
        """

        return np.array([vector4["x"], vector4["y"], vector4["z"], vector4["w"]])

    @staticmethod
    def array_to_vector4(arr: np.array) -> Dict[str, float]:
        """
        Convert a numpy array to a Vector4.

        :param arr: The numpy array.

        :return A Vector4, e.g. `{"x": 0, "y": 0, "z": 0, "w": 0}`
        """

        return {"x": arr[0], "y": arr[1], "z": arr[2], "w": arr[3]}

    @staticmethod
    def color_to_array(color: Dict[str, float]) -> np.array:
        """
        Convert a RGB Color to a numpy array.

        :param color: The Color object, e.g. `{"r": 0, "g": 0, "b": 0, "a": 1}`

        :return A numpy array.
        """

        return np.array([round(color["r"] * 255), round(color["g"] * 255), round(color["b"] * 255)])

    @staticmethod
    def array_to_color(arr: np.array) -> Dict[str, float]:
        """
        Convert a numpy array to a RGBA Color. If no A value is supplied it will default to 1.

        :param arr: The array.

        :return A Color, e.g. `{"r": 0, "g": 0, "b": 0, "a": 1}`
        """

        return {"r": arr[0], "g": arr[1], "b": arr[2], "a": 1 if len(arr) == 3 else arr[3]}

    @staticmethod
    def get_random_point_in_circle(center: np.array, radius: float) -> np.array:
        """
        Get a random point in a circle, defined by a center and radius.

        :param center: The center of the circle.
        :param radius: The radius of the circle.

        :return A numpy array. The y value (`arr[1]`) is always 0.
        """

        alpha = 2 * math.pi * random.random()
        r = radius * math.sqrt(random.random())
        x = r * math.cos(alpha) + center[0]
        z = r * math.sin(alpha) + center[2]

        return np.array([x, 0, z])

    @staticmethod
    def get_magnitude(vector3: Dict[str, float]) -> float:
        """
        Get the magnitude of a Vector3.

        :param vector3: The Vector3 object, e.g. `{"x": 0, "y": 0, "z": 0}`

        :return The vector magnitude.
        """

        return np.linalg.norm(TDWUtils.vector3_to_array(vector3))

    @staticmethod
    def extend_line(p0: np.array, p1: np.array, d: float, clamp_y=True) -> np.array:
        """
        Extend the line defined by p0 to p1 by distance d. Clamps the y value to 0.

        :param p0: The origin.
        :param p1: The second point.
        :param d: The distance of which the line is to be extended.
        :param clamp_y: Clamp the y value to 0.

        :return: The position at distance d.
        """

        if clamp_y:
            p0[1] = 0
            p1[1] = 0

        # Get the distance between the two points.
        d0 = distance.euclidean(p0, p1)
        # Get the total distance.
        d_total = d0 + d

        return p1 + ((p1 - p0) * d_total)

    @staticmethod
    def get_distance(vector3_0: Dict[str, float], vector3_1: Dict[str, float]) -> float:
        """
        Calculate the distance between two Vector3 (e.g. `{"x": 0, "y": 0, "z": 0}`) objects.

        :param vector3_0: The first Vector3.
        :param vector3_1: The second Vector3.

        :return The distance.
        """

        return distance.euclidean(TDWUtils.vector3_to_array(vector3_0), TDWUtils.vector3_to_array(vector3_1))

    @staticmethod
    def get_box(width: int, length: int) -> List[Dict[str, int]]:
        """
        Returns a list of x,y positions that can be used to create a box with the `create_exterior_walls` command.
        :param width: The width of the box.
        :param length: The length of the box.

        :return The box as represented by a list of `{"x": x, "y": y}` dictionaries.
        """

        box = []
        for x in range(width):
            for y in range(length):
                if x == 0 or x == width - 1 or y == 0 or y == length - 1:
                    box.append({"x": x, "y": y})
        return box

    @staticmethod
    def get_vector3(x, y, z) -> Dict[str, float]:
        """
        :param x: The x value.
        :param y: The y value.
        :param z: The z value.

        :return: A Vector3: {"x": x, "y", y, "z": z}
        """

        return {"x": x, "y": y, "z": z}

    @staticmethod
    def create_empty_room(width: int, length: int) -> dict:
        """
        :param width: The width of the room.
        :param length: The length of the room.

        :return: A `create_exterior_walls` command that creates a box with dimensions (width, length).
        """

        return {"$type": "create_exterior_walls", "walls": TDWUtils.get_box(width, length)}

    @staticmethod
    def create_room_from_image(filepath: str, exterior_color=(255, 0, 0), interior_color=(0, 0, 0)) -> List[dict]:
        """
        Load a .png file from the disk and use it to create a room. Each pixel on the image is a grid point.

        :param filepath: The absolute filepath to the image.
        :param exterior_color: The color on the image marking exterior walls (default=red).
        :param interior_color: The color on the image marking interior walls (default=black).

        :return: A list of commands: The first creates the exterior walls, and the second creates the interior walls.
        """

        exterior_walls = []
        interior_walls = []

        # Read the image.
        img = Image.open(filepath)
        pixels = img.load()
        col, row = img.size

        # Read each pixel as a grid point.
        for i in range(row):
            for j in range(col):
                pixel = pixels[i, j]
                if len(pixel) == 4:
                    pixel = (pixel[0], pixel[1], pixel[2])
                if pixel == exterior_color:
                    exterior_walls.append({"x": i, "y": col - j})
                elif pixel == interior_color:
                    interior_walls.append({"x": i, "y": col - j})

        return [{"$type": "create_exterior_walls",
                 "walls": exterior_walls},
                {"$type": "create_interior_walls",
                 "walls": interior_walls}]

    @staticmethod
    def save_images(images: Images, filename: str, output_directory="dist", resize_to=None, append_pass: bool = True) -> None:
        """
        Save each image in the Images object.
        The name of the image will be: pass_filename.extension, e.g.: `"0000"` -> `depth_0000.png`
        The images object includes the pass and extension information.

        :param images: The Images object. Contains each capture pass plus metadata.
        :param output_directory: The directory to write images to.
        :param filename: The filename of each image, minus the extension. The image pass will be appended as a prefix.
        :param resize_to: Specify a (width, height) tuple to resize the images to. This is slower than saving as-is.
        :param append_pass: If false, the image pass will _not_ be appended to the filename as a prefix, e.g.: `"0000"`: -> "`0000.jpg"`
        """

        if not os.path.isdir(output_directory):
            os.makedirs(output_directory)

        for i in range(images.get_num_passes()):
            if append_pass:
                fi = images.get_pass_mask(i)[1:] + "_" + filename + "." + images.get_extension(i)
            else:
                fi = filename + "." + images.get_extension(i)

            if resize_to:
                TDWUtils.get_pil_image(images, i).resize((resize_to[0], resize_to[1]), Image.LANCZOS)\
                    .save(os.path.join(output_directory, fi))
            else:
                pass_mask = images.get_pass_mask(i)
                path = os.path.join(output_directory, fi)
                # The depth passes aren't png files, so we need to convert them.
                if pass_mask == "_depth" or pass_mask == "_depth_simple":
                    # Save the image.
                    Image.fromarray(TDWUtils.get_shaped_depth_pass(images=images, index=i)).save(path)
                # Every other pass can be saved directly to disk.
                else:
                    with open(path, "wb") as f:
                        f.write(images.get_image(i))

    @staticmethod
    def get_shaped_depth_pass(images: Images, index: int) -> np.array:
        """
        The `_depth` and `_depth_simple` passes are a 1D array of RGB values, as oppposed to a png or jpg like every other pass.
        This function reshapes the array into a 2D array of RGB values.

        :param images: The `Images` output data.
        :param index: The index in `Images` of the depth pass. See: `Images.get_pass_mask()`.

        :return: A reshaped depth pass. Shape is: `(height, width, 3)`.
        """

        return np.flip(np.reshape(images.get_image(index), (images.get_height(), images.get_width(), 3)), 0)

    @staticmethod
    def zero_padding(integer: int, width=4) -> str:
        """
        :param integer: The integer being converted.
        :param width: The total number of digits in the string. If integer == 3 and width == 4, output is: "0003".

        :return A string representation of an integer padded with zeroes, e.g. converts `3` to `"0003"`.
        """

        return str(integer).zfill(width)

    @staticmethod
    def get_pil_image(images: Images, index: int) -> Image:
        """
        Converts Images output data to a PIL Image object.
        Use this function to read and analyze an image in memory.
        Do NOT use this function to save image data to disk; `save_image` is much faster.

        :param images: Images data from the build.
        :param index: The index of the image in Images.get_image

        :return A PIL image.
        """

        pass_mask = images.get_pass_mask(index)
        if pass_mask == "_depth" or pass_mask == "_depth_simple":
            return Image.fromarray(TDWUtils.get_shaped_depth_pass(images=images, index=index))
        else:
            return Image.open(io.BytesIO(images.get_image(index)))

    @staticmethod
    def get_random_position_on_nav_mesh(c: Controller, width: float, length: float, x_e=0, z_e=0, bake=True, rng=random.uniform) -> Tuple[float, float, float]:
        """
        Returns a random position on a NavMesh.

        :param c: The controller.
        :param width: The width of the environment.
        :param length: The length of the environment.
        :param bake: If true, send bake_nav_mesh.
        :param rng: Random number generator.
        :param x_e: The x position of the environment.
        :param z_e: The z position of the environment.

        :return The coordinates as a tuple `(x, y, z)`
        """

        if bake:
            c.communicate({'$type': 'bake_nav_mesh'})

        # Try to find a valid position on the NavMesh.
        is_on = False
        x, y, z = (0, 0, 0)
        while not is_on:
            # Get a random position.
            x = rng(-width / 2, width / 2) + x_e
            z = rng(-length / 2, length / 2) + z_e
            resp = c.communicate(
                {'$type': 'send_is_on_nav_mesh',
                 'position': {'x': x, 'y': 0, 'z': z},
                 'max_distance': 4.0
                 })
            answer = IsOnNavMesh(resp[0])
            is_on = answer.get_is_on()
            x, y, z = answer.get_position()
        return x, y, z

    @staticmethod
    def set_visual_material(c: Controller, substructure: List[dict], object_id: int, material: str, quality="med") -> List[dict]:
        """
        :param c: The controller.
        :param substructure: The metadata substructure of the object.
        :param object_id: The ID of the object in the scene.
        :param material: The name of the new material.
        :param quality: The quality of the material.

        :return A list of commands to set ALL visual materials on an object to a single material.
        """

        commands = []
        for sub_object in substructure:
            for i in range(len(sub_object["materials"])):
                commands.extend([c.get_add_material(material, library="materials_" + quality + ".json"),
                                 {"$type": "set_visual_material",
                                  "id": object_id,
                                  "material_name": material,
                                  "object_name": sub_object["name"],
                                  "material_index": i}])
        return commands

    @staticmethod
    def get_depth_values(image: np.array, depth_pass: str = "_depth", width: int = 256, height: int = 256, near_plane: float = 0.1, far_plane: float = 100) -> np.array:
        """
        Get the depth values of each pixel in a _depth image pass.
        The far plane is hardcoded as 100. The near plane is hardcoded as 0.1.
        (This is due to how the depth shader is implemented.)

        :param image: The image pass as a numpy array.
        :param depth_pass: The type of depth pass. This determines how the values are decoded. Options: `"_depth"`, `"_depth_simple"`.
        :param width: The width of the screen in pixels. See output data `Images.get_width()`.
        :param height: The height of the screen in pixels. See output data `Images.get_height()`.
        :param near_plane: The near clipping plane. See command `set_camera_clipping_planes`. The default value in this function is the default value of the near clipping plane.
        :param far_plane: The far clipping plane. See command `set_camera_clipping_planes`. The default value in this function is the default value of the far clipping plane.

        :return An array of depth values.
        """

        # Convert the image to a 2D image array.
        image = np.flip(np.reshape(image, (height, width, 3)), 0)
        if depth_pass == "_depth":
            depth_values = np.array((image[:, :, 0] + image[:, :, 1] / 256.0 + image[:, :, 2] / (256.0 ** 2)))
        elif depth_pass == "_depth_simple":
            depth_values = image[:, :, 0] / 256.0
        else:
            raise Exception(f"Invalid depth pass: {depth_pass}")
        # Un-normalize the depth values.
        return (depth_values * ((far_plane - near_plane) / 256.0)).astype(np.float32)

    @staticmethod
    def get_point_cloud(depth, camera_matrix: Union[np.array, tuple], vfov: float = 54.43222, filename: str = None, near_plane: float = 0.1, far_plane: float = 100) -> np.array:
        """
        Create a point cloud from an numpy array of depth values.

        :param depth: Depth values converted from a depth pass. See: `TDWUtils.get_depth_values()`
        :param camera_matrix: The camera matrix as a tuple or numpy array. See: [`send_camera_matrices`](https://github.com/threedworld-mit/tdw/blob/master/Documentation/api/command_api.md#send_camera_matrices).
        :param vfov: The field of view. See: [`set_field_of_view`](https://github.com/threedworld-mit/tdw/blob/master/Documentation/api/command_api.md#set_field_of_view)
        :param filename: If not None, the point cloud data will be written to this file.
        :param near_plane: The near clipping plane. See command `set_camera_clipping_planes`. The default value in this function is the default value of the near clipping plane.
        :param far_plane: The far clipping plane. See command `set_camera_clipping_planes`. The default value in this function is the default value of the far clipping plane.

        :return: An point cloud as a numpy array of `[x, y, z]` coordinates.
        """

        if isinstance(camera_matrix, tuple):
            camera_matrix = np.array(camera_matrix)
        camera_matrix = np.linalg.inv(camera_matrix.reshape((4, 4)))

        # Different from real-world camera coordinate system.
        # OpenGL uses negative z axis as the camera front direction.
        # x axes are same, hence y axis is reversed as well.
        # Source: https://learnopengl.com/Getting-started/Camera
        rot = np.array([[1, 0, 0, 0],
                        [0, -1, 0, 0],
                        [0, 0, -1, 0],
                        [0, 0, 0, 1]])
        camera_matrix = np.dot(camera_matrix, rot)

        # Cache some calculations we'll need to use every time.
        if TDWUtils.__HEIGHT != depth.shape[0] or TDWUtils.__WIDTH != depth.shape[1]:
            TDWUtils.__HEIGHT = depth.shape[0]
            TDWUtils.__WIDTH = depth.shape[1]

            img_pixs = np.mgrid[0: depth.shape[0], 0: depth.shape[1]].reshape(2, -1)
            # Swap (v, u) into (u, v).
            img_pixs[[0, 1], :] = img_pixs[[1, 0], :]
            img_pix_ones = np.concatenate((img_pixs, np.ones((1, img_pixs.shape[1]))))

            # Calculate the intrinsic matrix from vertical_fov.
            # Motice that hfov and vfov are different if height != width
            # We can also get the intrinsic matrix from opengl's perspective matrix.
            # http://kgeorge.github.io/2014/03/08/calculating-opengl-perspective-matrix-from-opencv-intrinsic-matrix
            vfov = vfov / 180.0 * np.pi
            tan_half_vfov = np.tan(vfov / 2.0)
            tan_half_hfov = tan_half_vfov * TDWUtils.__WIDTH / float(TDWUtils.__HEIGHT)
            fx = TDWUtils.__WIDTH / 2.0 / tan_half_hfov  # focal length in pixel space
            fy = TDWUtils.__HEIGHT / 2.0 / tan_half_vfov
            intrinsics = np.array([[fx, 0, TDWUtils.__WIDTH / 2.0],
                                   [0, fy, TDWUtils.__HEIGHT / 2.0],
                                   [0, 0, 1]])
            img_inv = np.linalg.inv(intrinsics[:3, :3])
            TDWUtils.__CAM_TO_IMG_MAT = np.dot(img_inv, img_pix_ones)

        points_in_cam = np.multiply(TDWUtils.__CAM_TO_IMG_MAT, depth.reshape(-1))
        points_in_cam = np.concatenate((points_in_cam, np.ones((1, points_in_cam.shape[1]))), axis=0)
        points_in_world = np.dot(camera_matrix, points_in_cam)
        points_in_world = points_in_world[:3, :].reshape(3, TDWUtils.__WIDTH, TDWUtils.__HEIGHT)
        points_in_cam = points_in_cam[:3, :].reshape(3, TDWUtils.__WIDTH, TDWUtils.__HEIGHT)
        if filename is not None:
            f = open(filename, 'w')
            for i in range(points_in_world.shape[1]):
                for j in range(points_in_world.shape[2]):
                    if points_in_cam[2, i, j] < (far_plane - near_plane):
                        f.write(f'{points_in_world[0, i, j]};{points_in_world[1, i, j]};{points_in_world[2, i, j]}\n')
        return points_in_world

    @staticmethod
    def create_avatar(avatar_type="A_Img_Caps_Kinematic", avatar_id="a", position=None, look_at=None) -> List[dict]:
        """
        This is a wrapper for `create_avatar` and, optionally, `teleport_avatar_to` and `look_at_position`.

        :param avatar_type: The type of avatar.
        :param avatar_id: The avatar ID.
        :param position: The position of the avatar. If this is None, the avatar won't teleport.
        :param look_at: If this isn't None, the avatar will look at this position.

        :return A list of commands to create theavatar.
        """

        # Create the avatar.
        commands = [{"$type": "create_avatar",
                     "type": avatar_type,
                     "id": avatar_id}]

        # Teleport the avatar.
        if position:
            commands.append({"$type": "teleport_avatar_to",
                             "avatar_id": avatar_id,
                             "position": position})
        if look_at:
            commands.append({"$type": "look_at_position",
                             "avatar_id": avatar_id,
                             "position": look_at})
        return commands

    @staticmethod
    def get_unit_scale(record: ModelRecord) -> float:
        """
        :param record: The model record.

        :return The scale factor required to scale a model to 1 meter "unit scale".
        """

        bounds = record.bounds

        # Get the "unit scale" of the object.
        try:
            s = 1 / max(
                bounds['top']['y'] - bounds['bottom']['y'],
                bounds['front']['z'] - bounds['back']['z'],
                bounds['right']['x'] - bounds['left']['x'])
            return s
        except ZeroDivisionError:
            return 1

    @staticmethod
    def validate_amazon_s3() -> bool:
        """
        Validate that your local Amazon S3 credentials are set up correctly.

        :return True if everything is OK.
        """

        config_path = Path.home().joinpath(".aws/config")
        new_config_path = not config_path.exists()
        # Generate a valid config file.
        if new_config_path:
            config_path.write_text("[default]\nregion = us-east-1\noutput = json")
            print(f"Generated a new config file: {config_path.resolve()}")
        try:
            session = boto3.Session(profile_name="tdw")
            s3 = session.resource("s3")
            s3.meta.client.head_object(Bucket='tdw-private', Key='models/windows/2018-2019.1/iron_box')
            return True
        except ProfileNotFound:
            print(f"ERROR! Your AWS credentials file is not set up correctly.")
            print("Your AWS credentials must have a [tdw] profile with valid keys.")
            return False
        except ClientError as e:
            print("ERROR! Could not access bucket tdw-private. Make sure you have the right permissions.")
            error_code = e.response['Error']['Code']
            print(e, error_code)
            return False

    @staticmethod
    def get_base64_flex_particle_forces(forces: list) -> str:
        """
        :param forces: The forces (see Flex documentation for how to arrange this array).

        :return: An array of Flex particle forces encoded in base64.
        """

        forces = np.array(forces, dtype=np.float32)
        return base64.b64encode(forces).decode()

    @staticmethod
    def color_to_hashable(color: Union[np.array, Tuple[int, int, int]]) -> int:
        """
        :param color: The color as an RGB array or tuple, where each value is between 0 and 255.

        :return: A hashable integer representation of the color array.
        """

        return (color[0] << 16) + (color[1] << 8) + color[2]

    @staticmethod
    def hashable_to_color(hashable: int) -> np.array:
        """
        :param hashable: A hashable integer representing an RGB color.

        :return: A color as a numpy array of integers between 0 and 255: `[r, g, b]`
        """

        return np.array([(hashable >> 16) & 255, (hashable >> 8) & 255, hashable & 255], dtype=int)

    @staticmethod
    def get_bounds_dict(bounds: Bounds, index: int) -> Dict[str, np.array]:
        """
        :param bounds: Bounds output data.
        :param index: The index in `bounds` of the target object.

        :return: A dictionary of the bounds. Key = the name of the position. Value = the position as a numpy array.
        """

        return {"top": np.array(bounds.get_top(index)),
                "bottom": np.array(bounds.get_bottom(index)),
                "left": np.array(bounds.get_left(index)),
                "right": np.array(bounds.get_right(index)),
                "front": np.array(bounds.get_front(index)),
                "back": np.array(bounds.get_back(index)),
                "center": np.array(bounds.get_center(index))}

    @staticmethod
    def get_bounds_extents(bounds: Union[Bounds, Dict[str, Dict[str, float]]], index: int = 0) -> np.array:
        """
        :param bounds: Bounds output data or cached bounds data from a record (`record.bounds`).
        :param index: The index in `bounds` of the target object. Ignored if `bounds` is a dictionary.

        :return: The width (left to right), height (top to bottom), and length (front to back), of the bounds as a numpy array.
        """

        if isinstance(bounds, Bounds):
            return np.array([np.linalg.norm(np.array(bounds.get_left(index)) - np.array(bounds.get_right(index))),
                             np.linalg.norm(np.array(bounds.get_top(index)) - np.array(bounds.get_bottom(index))),
                             np.linalg.norm(np.array(bounds.get_front(index)) - np.array(bounds.get_back(index)))])
        elif isinstance(bounds, dict):
            return np.array([np.linalg.norm(TDWUtils.vector3_to_array(bounds["left"]) - TDWUtils.vector3_to_array(bounds["right"])),
                             np.linalg.norm(TDWUtils.vector3_to_array(bounds["top"]) - TDWUtils.vector3_to_array(bounds["bottom"])),
                             np.linalg.norm(TDWUtils.vector3_to_array(bounds["front"]) - TDWUtils.vector3_to_array(bounds["back"]))])
        else:
            raise Exception(f"Invalid bounds data: {bounds}")

    @staticmethod
    def get_closest_position_in_bounds(origin: np.array, bounds: Bounds, index: int) -> np.array:
        """
        :param origin: The origin from which the distance is calculated.
        :param bounds: Bounds output data.
        :param index: The index in `bounds` of the target object.

        :return: The position on the object bounds that is closest to `origin`.
        """

        object_bounds = TDWUtils.get_bounds_dict(bounds=bounds, index=index)

        # Get the closest point on the bounds.
        min_destination = ""
        min_distance = 10000
        for p in object_bounds:
            d = np.linalg.norm(origin - object_bounds[p])
            if d < min_distance:
                min_distance = d
                min_destination = p
        return object_bounds[min_destination]

    @staticmethod
    def get_angle(forward: np.array, origin: np.array, position: np.array) -> float:
        """
          :param position: The target position.
          :param origin: The origin position of the directional vector.
          :param forward: The forward directional vector.

          :return: The angle in degrees between `forward` and the direction vector from `origin` to `position`.
          """

        # Get the normalized directional vector to the target position.
        p0 = np.array([origin[0], origin[2]])
        p1 = np.array([position[0], position[2]])
        d = p1 - p0
        d = d / np.linalg.norm(d)
        f = np.array([forward[0], forward[2]])

        dot = f[0] * d[0] + f[1] * d[1]
        det = f[0] * d[1] - f[1] * d[0]
        angle = np.arctan2(det, dot)
        angle = np.rad2deg(angle)
        return angle

    @staticmethod
    def get_angle_between(v1: np.array, v2: np.array) -> float:
        """
        :param v1: The first directional vector.
        :param v2: The second directional vector.

        :return: The angle in degrees between two directional vectors.
        """

        ang1 = np.arctan2(v1[2], v1[0])
        ang2 = np.arctan2(v2[2], v2[0])

        return np.rad2deg((ang1 - ang2) % (2 * np.pi))

    @staticmethod
    def rotate_position_around(position: np.array, angle: float, origin: np.array = None) -> np.array:
        """
        Rotate a position by a given angle around a given origin.

        :param origin: The origin position.  If None, the origin is `[0, 0, 0]`
        :param position: The point being rotated.
        :param angle: The angle in degrees.

        :return: The rotated position.
        """

        if origin is None:
            origin = np.array([0, 0, 0])

        radians = np.deg2rad(angle)
        x, y = position[0], position[2]
        offset_x, offset_y = origin[0], origin[2]
        adjusted_x = (x - offset_x)
        adjusted_y = (y - offset_y)
        cos_rad = np.cos(radians)
        sin_rad = np.sin(radians)
        qx = offset_x + cos_rad * adjusted_x + sin_rad * adjusted_y
        qy = offset_y + -sin_rad * adjusted_x + cos_rad * adjusted_y

        return np.array([qx, position[1], qy])

    @staticmethod
    def euler_angles_to_rpy(euler_angles: np.array) -> np.array:
        """
        Convert Euler angles to ROS RPY angles.

        :param euler_angles: A numpy array: `[x, y, z]` Euler angles in degrees.

        :return: A numpy array: `[r, p, y]` angles in radians.
        """

        # Source: https://github.com/Unity-Technologies/URDF-Importer/blob/c41208565419b04907496baa93ad1b675d41dc20/com.unity.robotics.urdf-importer/Runtime/Extensions/TransformExtensions.cs#L85-L92
        return np.radians(np.array([-euler_angles[2], euler_angles[0], -euler_angles[1]]))

    @staticmethod
    def bytes_to_megabytes(b: int) -> float:
        """
        :param b: A quantity of bytes.

        :return: A quantity of megabytes.
        """

        return b / (1 << 20)

    @staticmethod
    def get_circle_mask(shape: Tuple[int, int], row: int, column: int, radius: int) -> np.array:
        """
        Get elements in an array within a circle.

        :param shape: The shape of the source array as (rows, columns).
        :param row: The row (axis 0) of the center of the circle.
        :param column: The column (axis 1) of the circle.
        :param radius: The radius of the circle in indices.

        :return: A boolean array with shape `shape`. Elements that are True are within the circle.
        """

        # Source: https://www.semicolonworld.com/question/44279/how-to-apply-a-disc-shaped-mask-to-a-numpy-array
        nx, ny = shape
        oy, ox = np.ogrid[-row:nx - row, -column:ny - column]
        return ox * ox + oy * oy <= radius * radius



from typing import List, Dict
from tdw.output_data import OutputData, StaticRobot
from tdw.robot_data.joint_static import JointStatic
from tdw.robot_data.non_moving import NonMoving


class RobotStatic:
    """
    Static data for a robot that won't change due to physics (such as the joint IDs, segmentation colors, etc.)
    """

    def __init__(self, robot_id: int, resp: List[bytes]):
        """
        :param resp: The response from the build, which we assume contains `Robot` output data.
        :param robot_id: The ID of this robot.
        """

        """:field
        The ID of the robot.
        """
        self.robot_id: int = robot_id
        """:field
        A dictionary of [Static robot joint data](joint_static.md) for each joint. Key = The ID of the joint.
        """
        self.joints: Dict[int, JointStatic] = dict()
        """:field
        A dictionary of joint names. Key = The name of the joint. Value = The joint ID.
        """
        self.joint_ids_by_name: Dict[str, int] = dict()
        """:field
        A dictionary of [Static data for non-moving parts](non_moving.md) for each non-moving part. Key = The ID of the part.
        """
        self.non_moving: Dict[int, NonMoving] = dict()
        """:field
        A list of joint IDs and non-moving body part IDs.
        """
        self.body_parts: List[int] = list()
        """:field
        If True, the robot is immovable.
        """
        self.immovable: bool = False
        for i in range(len(resp) - 1):
            r_id = OutputData.get_data_type_id(resp[i])
            if r_id == "srob":
                static_robot: StaticRobot = StaticRobot(resp[i])
                if static_robot.get_id() == robot_id:
                    for j in range(static_robot.get_num_joints()):
                        joint = JointStatic(static_robot=static_robot, joint_index=j)
                        self.joints[joint.joint_id] = joint
                        self.joint_ids_by_name[joint.name] = joint.joint_id
                        if joint.root:
                            self.immovable = joint.immovable
                    for j in range(static_robot.get_num_non_moving()):
                        non_moving = NonMoving(static_robot=static_robot, index=j)
                        self.non_moving[non_moving.object_id] = non_moving
        self.body_parts: List[int] = list(self.joints.keys())
        self.body_parts.extend(self.non_moving.keys())



from enum import Enum
from platform import system
from typing import List, Union, Optional
from tdw.controller import Controller
from tdw.librarian import ModelLibrarian, ModelRecord
from tdw.add_ons.add_on import AddOn
from tdw.add_ons.model_verifier.model_tests.model_test import ModelTest
from tdw.add_ons.model_verifier.model_tests.model_report import ModelReport
from tdw.add_ons.model_verifier.model_tests.missing_materials import MissingMaterials
from tdw.add_ons.model_verifier.model_tests.physics_quality import PhysicsQuality


class _Test(Enum):
    """
    A type of test.
    """

    model_report = 1
    missing_materials = 2
    physics_quality = 4


class ModelVerifier(AddOn):
    """
    Run tests on an object model.
    """

    def __init__(self):
        super().__init__()
        self._tests: List[_Test] = list()
        self._record: Optional[ModelRecord] = None
        self._current_test: Optional[ModelTest] = None
        self._object_id: int = -1
        """:field
        A list of reports from the test.
        """
        self.reports: List[str] = list()
        """:field
        If True, the tests are done.
        """
        self.done: bool = False

    def get_initialization_commands(self) -> List[dict]:
        return [{"$type": "create_empty_environment"},
                {"$type": "set_render_quality",
                 "render_quality": 0},
                {"$type": "set_post_process",
                 "value": False},
                {"$type": "simulate_physics",
                 "value": False},
                {"$type": "create_avatar",
                 "id": "a",
                 "type": "A_Img_Caps_Kinematic"},
                {"$type": "teleport_avatar_to",
                 "position": {"x": 1.75, "y": 0.5, "z": 0}},
                {"$type": "look_at_position",
                 "position": {"x": 0, "y": 0.5, "z": 0}}]

    def set_tests(self, name: str, model_report: bool, missing_materials: bool, physics_quality: bool,
                  source: Union[ModelLibrarian, ModelRecord, str] = None) -> None:
        """
        Start new tests for the model. Only call this if there isn't currently a test running.

        :param name: The name of the model.
        :param source: The source of the model. If None: Get the record corresponding to `name` in `models_core.json`. If `ModelLibrarian`: Get the record corresponding to `name` in this library. If `ModelRecord`: Use this record. If `str`: Create a dummy record with name `name` and URL `source`.
        :param model_report: If True, run a basic test on the model.
        :param missing_materials: If True, test the model for any missing materials.
        :param physics_quality: If True, test the extent to which the colliders geometry matches the rendered geometry.
        """

        if self._current_test is not None:
            raise Exception(f"Cannot set new tests because we're still testing {self._record.name}")
        self._tests = list()
        self.done = False
        self.reports.clear()
        if model_report:
            self._tests.append(_Test.model_report)
        if missing_materials:
            self._tests.append(_Test.missing_materials)
        if physics_quality:
            self._tests.append(_Test.physics_quality)
        # This is the name of a model in models_core.json
        if source is None:
            if "models_core.json" not in Controller.MODEL_LIBRARIANS:
                Controller.MODEL_LIBRARIANS["models_core.json"] = ModelLibrarian()
            self._record = Controller.MODEL_LIBRARIANS["models_core.json"].get_record(name)
        # This is the name of a record in a specified library.
        elif isinstance(source, ModelLibrarian):
            self._record = source.get_record(name)
        # This is a model record.
        elif isinstance(source, ModelRecord):
            self._record = source
        # This is a URL.
        elif isinstance(source, str):
            self._record = ModelRecord()
            self._record.name = name
            self._record.urls[system()] = source
        else:
            raise TypeError("Invalid type for source: " + source)
        self._next_test()

    def _next_test(self) -> None:
        """
        Start the next test, if any.
        """

        if len(self._tests) == 0:
            self._current_test = None
            self._record = None
            return
        t: _Test = self._tests.pop(0)
        if t == _Test.model_report:
            self._current_test = ModelReport(record=self._record)
        elif t == _Test.physics_quality:
            self._current_test = PhysicsQuality(record=self._record)
        elif t == _Test.missing_materials:
            self._current_test = MissingMaterials(record=self._record)
        else:
            raise Exception(f"Test not defined: {t}")
        self.commands.extend(self._current_test.start())

    def on_send(self, resp: List[bytes]) -> None:
        if self._current_test is None:
            self.done = True
            return
        # Get test commands.
        self.commands.extend(self._current_test.on_send(resp=resp))
        # If this test is done, append the reports and try to start a new test.
        if self._current_test.done:
            self.reports.extend(self._current_test.reports)
            self._next_test()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: FBOutput

class MachineType(object):
    none = 1
    motor = 2
    hinge = 3
    spring = 4
    light = 5
    prismatic_joint = 6




import base64
from pathlib import Path
from typing import Union
import wave
import numpy as np
from tdw.audio_constants import SAMPLE_RATE, CHANNELS, SAMPLE_WIDTH


class Base64Sound:
    """
    A sound encoded as a base64 string.
    """

    def __init__(self, snd: np.array):
        """
        :param snd: The sound byte array.
        """

        """:field
        Byte data of the sound.
        """
        self.bytes: bytes = bytes(np.array(snd * 32767, dtype='int16'))
        """:field
        A base64 string of the sound. Send this to the build.
        """
        self.wav_str = base64.b64encode(self.bytes).decode('utf-8')
        """:field
        The length of the byte array.
        """
        self.length: int = len(self.bytes)

    def write(self, path: Union[str, Path]) -> None:
        """
        Write audio to disk.

        :param path: The path to the .wav file.
        """

        if isinstance(path, Path):
            path = str(path.resolve())
        w = wave.Wave_write(f=path)
        w.setnchannels(CHANNELS)
        w.setnframes(self.length)
        w.setframerate(SAMPLE_RATE)
        w.setsampwidth(SAMPLE_WIDTH)
        w.writeframes(self.bytes)
